'''These are functions that generate various coverage vectors.
This module exports:
build_hitsclip_vectors
hitsclip_vectors_2_bg
'''
import pysam

from clippyl.ome_dict_io import SiteTuple, OmeDict, dump2bg

class build_hitsclip_vectors():
    '''Generate basewise counts for 1D, cleavage and raw coverage from the 
    set of alignments mapped to the region defined by ome_coords.
    This function parses the list of dictionaries generated by the pysam
    indexed bam file handle.
    Note: if the cleaved_readid_db_conn is set to None then it is assumed
    that all the reads in the sam file are adapter-clipped
    '''
    
    def __init__(self):
        
        self.stat_dict = {}
        self.stat_dict['n_of_reads'] = 0
        self.stat_dict['n_of_strand_matched_reads'] = 0
        self.stat_dict['n_of_unique_aligns'] = 0 #gleaned from sam XT opt field
        self.stat_dict['n_of_adapter_clipped_reads'] = 0
        
        self.stat_dict['n_of_nt_covered'] = 0
        self.stat_dict['n_of_nt_termini'] = 0
        self.stat_dict['n_of_oneD_operations'] = 0
        
        return
    
    def __call__(self, ome_coords,
                       pysam_bam_file_conn,
                       cleaved_readid_db_conn = None,
                       uniq_only = True,
                       oneD_rate_mode = True,
                       rate_cutoff = 15,
                       cleav_rate_mode = False,
                       stranded = True):
        
        reference, start, end, strand = ome_coords
        
        # fetch the bam-encoded alignment data for the region
        # delineated by the ome_coords.
        # note that it will also return reads that are only
        # partially overlapping with the region
        fetched_alignments = []
        i = pysam_bam_file_conn.fetch( reference,
                                       start,
                                       end )
        for alignment in i:
            # object class returned by fetch is pysam.AlignedSegment
            # attributes are listed here:
            #https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment
            d = {}
            d['QNAME'] = alignment.qname
            d['POS'] = alignment.pos
            d['AEND'] = alignment.aend
            d['CIGAR'] = alignment.cigar
            d['IS_REVERSE'] = alignment.is_reverse
            if alignment.is_reverse:
                d['strand'] = '-'
            else:
                d['strand'] = '+'
            d['OPT_XT'] = alignment.opt('XT')
            fetched_alignments.append(d)
        
        self.stat_dict['n_of_reads'] = len(fetched_alignments)
        
        # instantiate OmeDict, which will aggregate and then output the
        # relevant basewise data as it is gleaned from the alignment data
        raw_cover_cd = OmeDict()
        cleavage_cd = OmeDict()
        oneD_cd = OmeDict()
        
        for d in fetched_alignments:
            
            if strand == d['strand']:
                pass
            else:
                continue
            
            self.stat_dict['n_of_strand_matched_reads'] += 1
            
            #identify uniquely aligned reads by accessing the sam optdict
            if d['OPT_XT'] == 'U':
                self.stat_dict['n_of_unique_aligns'] += 1
            elif uniq_only:
                continue
            else:
                pass
            
            ####
            #store basewise raw coverage data in OmeDict
            raw_cover_st = SiteTuple(reference,
                                     start = d['POS'],
                                     end = d['AEND'],
                                     strand = d['strand'],
                                     data = 1)
            raw_cover_cd.add_site_tuple( raw_cover_st )
            
            ####
            #store basewise cleavage data in OmeDict (if present)
            if cleaved_readid_db_conn:
                if cleaved_readid_db_conn.readid_lookup(d['QNAME']):
                    self.stat_dict['n_of_adapter_clipped_reads'] += 1
                    
                    #add left cleavage site
                    left_terminus_st = SiteTuple(reference,
                                                 start = d['POS'],
                                                 end = d['POS'] + 1,
                                                 strand = d['strand'],
                                                 data = 1)
                    cleavage_cd.add_site_tuple( left_terminus_st )
                    #add right cleavage site
                    right_terminus_st = SiteTuple(reference,
                                                  start = d['AEND'] - 1,
                                                  end = d['AEND'],
                                                  strand = d['strand'],
                                                  data = 1)
                    cleavage_cd.add_site_tuple( right_terminus_st )
                else:
                    pass
            else:
                #No clipped read database provided, therefore, assuming that 
                #all reads are assumed to be adapter-clipped
                
                #add left cleavage site
                left_terminus_st = SiteTuple(reference,
                                             start = d['POS'],
                                             end = d['POS'] + 1,
                                             strand = d['strand'],
                                             data = 1)
                cleavage_cd.add_site_tuple( left_terminus_st )
                #add right cleavage site
                right_terminus_st = SiteTuple(reference,
                                              start = d['AEND'] - 1,
                                              end = d['AEND'],
                                              strand = d['strand'],
                                              data = 1)
                cleavage_cd.add_site_tuple( right_terminus_st )
            
            ####
            # store 1D (aka single nucleotide deletion) data in OmeDict
            # note: pysam uses an integer key to refer to CIGAR operations
            # http://pysam.readthedocs.org/en/latest/api.html#pysam.AlignedRead.cigar
            # deletion operations are coded as a 2
            if (2, 1) in d['CIGAR']:
                # the postional index of the alignment
                # operation (on the reference) will be
                # stored as i
                i = -1
                for operation, length in d['CIGAR']:
                    if operation != 1:
                        i += length
                        if operation == 2 and length == 1:
                            oneD_st = SiteTuple(reference,
                                                start = d['POS'] + i,
                                                end = d['POS'] + i + 1,
                                                strand = d['strand'],
                                                data = 1)
                            oneD_cd.add_site_tuple(oneD_st)
                        else:
                            pass
        
        # output the data vectors. 
        # note that the get_data method reverses the bottom strand vectors by 
        # default (see orient_strands argument). Therefore the top and bottom 
        # strand coverage vectors have the same 5'->3' polarity.
        # also note that the stranded argument here can be toggled to accomodate
        # strand-agnostic sequencing libraries. However HITS-CLIP libraries are
        # typically stranded.
        if stranded:
            raw_cover_vec = raw_cover_cd.get_data(ome_coords)
            self.stat_dict['n_of_nt_covered'] = sum(raw_cover_vec)
            cleavage_vec = cleavage_cd.get_data(ome_coords)
            self.stat_dict['n_of_nt_termini'] = sum(cleavage_vec)
            oneD_vec = oneD_cd.get_data(ome_coords)
            self.stat_dict['n_of_oneD_operations'] = sum(oneD_vec)
        else:
            raw_cover_vec = raw_cover_cd.get_data(ome_coords, both_strands = True)
            self.stat_dict['n_of_nt_covered'] = sum(raw_cover_vec)
            cleavage_vec = cleavage_cd.get_data(ome_coords, both_strands = True)
            self.stat_dict['n_of_nt_termini'] = sum(cleavage_vec)
            oneD_vec = oneD_cd.get_data(ome_coords, both_strands = True)
            self.stat_dict['n_of_oneD_operations'] = sum(oneD_vec)
        
        if oneD_rate_mode == True:
            oneD_rate_vec = []
            for x, y in zip(oneD_vec, raw_cover_vec):
                if y < rate_cutoff:
                    oneD_rate_vec.append(0)
                else:
                    oneD_rate_vec.append(x/y)
            
            oneD_vec = oneD_rate_vec
        
        if cleav_rate_mode == True:
            cleav_rate_vec = []
            for x, y in zip(cleavage_vec, raw_cov_vec):
                if y < rate_cutoff:
                    cleav_rate_vec.append(0)
                else:
                    cleav_rate_vec.append(x/y)
            
            cleavage_vec = cleav_rate_vec
        
        vector_tuple = ( raw_cover_vec,
                         oneD_vec,
                         cleavage_vec )
        
        return vector_tuple

class hitsclip_vectors_2_bg():
    """Generate multiple bedgraph files containing multiple coverage 
    metrics relevant to hits-clip. Note: if the cleav_db_fp is set to 
    None then it is assumed that all the reads in the sam file are 
    adapter-clipped
    """
    
    def __init__(self):
        
        self.stat_dict = {}
        self.stat_dict['n_of_mapped_reads'] = 0
        self.stat_dict['n_of_unique_aligns'] = 0 #gleaned from sam XT opt field
        self.stat_dict['n_of_adapter_clipped_reads'] = 0
        
        self.stat_dict['n_of_nt_covered'] = 0
        self.stat_dict['n_of_nt_termini'] = 0
        self.stat_dict['n_of_oneD_operations'] = 0
        
        return
    
    def __call__(self, pysam_bam_file_conn,
                       sample_name,
                       bg_fh_top_strand_cov,
                       bg_fh_bot_strand_cov,
                       bg_fh_top_strand_term,
                       bg_fh_bot_strand_term,
                       bg_fh_top_strand_1D,
                       bg_fh_bot_strand_1D,
                       cleaved_readid_db_conn = None,
                       uniq_only = True,
                       oneD_rate_mode = True,
                       oneD_rate_cutoff = 15,
                       cleav_rate_mode = False,
                       cleav_rate_cutoff = 0,
                       max_chunk_size = 1000000):
        
        #see bedgraph definition at https://genome.ucsc.edu/goldenPath/help/bedgraph.html
        #note  bedGraphToBigWig can compress the file see http://genome.ucsc.edu/goldenPath/help/bigWig.html#Ex3
        generic_head = 'track type=bedGraph name="{0}" description="{1}{2}" visibility=full\n'
        bg_fh_top_strand_cov.write(generic_head.format('5\'--> coverage', sample_name, 'Top strand coverage'))
        bg_fh_bot_strand_cov.write(generic_head.format('<--3\' coverage', sample_name, 'Bottom strand coverage'))
        bg_fh_top_strand_term.write(generic_head.format('5\'--> termini', sample_name, 'Top strand termini'))
        bg_fh_bot_strand_term.write(generic_head.format('<--3\' coverage', sample_name, 'Bottom strand termini'))
        bg_fh_top_strand_1D.write(generic_head.format('5\'--> coverage', sample_name, 'Top strand single nucleotide deletions'))
        bg_fh_bot_strand_1D.write(generic_head.format('<--3\' coverage', sample_name, 'Bottom single nucleotide deletions'))
        
        
        # instantiate OmeDict, which will aggregate and then output the
        # relevant basewise data as it is gleaned from the alignment data
        raw_cover_cd = OmeDict()
        cleavage_cd = OmeDict()
        oneD_cd = OmeDict()
        
        chunk_size = 0 # an object to manage memory tracks with n_of_nt_covered
        
        # store reference lengths
        ref_length_d = {}
        for ref, ref_length in zip(pysam_bam_file_conn.references, pysam_bam_file_conn.lengths):
            ref_length_d[ref] = int(ref_length)
        
        for alignment in pysam_bam_file_conn.fetch():
            
            d = {}
            #TODO: verify that zero-based coord is returned
            d['QNAME'] = alignment.qname # read ID
            d['RNAME'] =  pysam_bam_file_conn.getrname(alignment.tid) # reference
            d['POS'] = alignment.pos
            d['AEND'] = alignment.aend
            # alignments that span multiple references won't be considered
            if d['AEND'] > ref_length_d[d['RNAME']]:
                continue
            d['CIGAR'] = alignment.cigar
            d['IS_REVERSE'] = alignment.is_reverse
            if alignment.is_reverse:
                d['strand'] = '-'
            else:
                d['strand'] = '+'
            d['OPT_XT'] = alignment.opt('XT')
            
            self.stat_dict['n_of_mapped_reads'] += 1
            
            #identify uniquely aligned reads by accessing the sam optdict
            if d['OPT_XT'] == 'U':
                self.stat_dict['n_of_unique_aligns'] += 1
            elif uniq_only:
                continue
            else:
                pass
            
            ####
            #store basewise raw coverage data in OmeDict
            raw_cover_st = SiteTuple(reference = d['RNAME'],
                                     start = d['POS'],
                                     end = d['AEND'],
                                     strand = d['strand'],
                                     data = 1)
            n = raw_cover_cd.add_site_tuple( raw_cover_st )
            self.stat_dict['n_of_nt_covered'] += n
            chunk_size += n
            
            ####
            #store basewise cleavage data in OmeDict (if present)
            if cleaved_readid_db_conn:
                if cleaved_readid_db_conn.readid_lookup(d['QNAME']):
                    self.stat_dict['n_of_adapter_clipped_reads'] += 1
                    
                    #add left cleavage site
                    left_terminus_st = SiteTuple(reference = d['RNAME'],
                                                 start = d['POS'],
                                                 end = d['POS'] + 1,
                                                 strand = d['strand'],
                                                 data = 1)
                    n = cleavage_cd.add_site_tuple( left_terminus_st )
                    
                    #add right cleavage site
                    right_terminus_st = SiteTuple(reference = d['RNAME'],
                                                  start = d['AEND'] - 1,
                                                  end = d['AEND'],
                                                  strand = d['strand'],
                                                  data = 1)
                    n = cleavage_cd.add_site_tuple( right_terminus_st )
                else:
                    pass
            else:
                #No clipped read database provided, therefore, assuming that 
                #all reads are adapter-clipped
                
                #add left cleavage site
                left_terminus_st = SiteTuple(reference = d['RNAME'],
                                             start = d['POS'],
                                             end = d['POS'] + 1,
                                             strand = d['strand'],
                                             data = 1)
                n = cleavage_cd.add_site_tuple( left_terminus_st )
                
                #add right cleavage site
                right_terminus_st = SiteTuple(reference = d['RNAME'],
                                              start = d['AEND'] - 1,
                                              end = d['AEND'],
                                              strand = d['strand'],
                                              data = 1)
                n = cleavage_cd.add_site_tuple( right_terminus_st )
            
            ####
            # store 1D (aka single nucleotide deletion) data in OmeDict
            # note: pysam uses an integer key to refer to CIGAR operations
            # http://pysam.readthedocs.org/en/latest/api.html#pysam.AlignedRead.cigar
            # deletion operations are coded as a 2
            if (2, 1) in d['CIGAR']:
                # the postional index of the alignment
                # operation (on the reference) will be
                # stored as i
                i = -1
                for operation, length in d['CIGAR']:
                    if operation != 1:
                        i += length
                        if operation == 2 and length == 1:
                            oneD_st = SiteTuple(reference = d['RNAME'],
                                                start = d['POS'] + i,
                                                end = d['POS'] + i + 1,
                                                strand = d['strand'],
                                                data = 1)
                            n = oneD_cd.add_site_tuple(oneD_st)
                        else:
                            pass
            
            if chunk_size >= max_chunk_size:
                print('chunk size exceeded, purge memory', chunk_size)
                #TODO:writeout all coords 5' of the most recent alignment and delete
                # the key-value pairs from memory. include a print statement that 
                # declares the writeout is occuring so that memory clean-up
                # can be verified by watching htop
                
                # output the data vectors. 
                # note that the get_data method reverses the bottom strand vectors by 
                # default (see orient_strands argument). Therefore the top and bottom 
                # strand coverage vectors have the same 5'->3' polarity.
                
                # NOTE: the coordinates of the current alignment comprises
                # the boundary coords. Everything upstream will be dumped
                # to file and deleted from memory.
                b_ref = d['RNAME'] # current ref is the boundary ref
                b_coord = d['POS'] # the boundary is the start coord
                
                print('purge cleav') #debugging
                t = dump2bg( cleavage_cd, 
                             bg_fh_top_strand_term, 
                             bg_fh_bot_strand_term, 
                             b_ref, # current boundary ref
                             b_coord, # current boundary coord
                             rate_mode = cleav_rate_mode, 
                             rate_cutoff = cleav_rate_cutoff,
                             rate_denom_cd = raw_cover_cd,
                              )
                #self.stat_dict['n_of_nt_termini'] += n
                
                print('purge 1D') #debugging
                t = dump2bg( oneD_cd, 
                             bg_fh_top_strand_1D, 
                             bg_fh_bot_strand_1D, 
                             b_ref, # current boundary ref
                             b_coord, # current boundary coord
                             rate_mode = oneD_rate_mode, 
                             rate_cutoff = oneD_rate_cutoff,
                             rate_denom_cd = raw_cover_cd,
                              )
                #self.stat_dict['n_of_oneD_operations'] += n
                
                # The raw coverage ome_dict must be dumped last because it 
                # serves as the denominator for the rate calculations for 
                # cleavage and oneD
                print('purge raw cov') #debugging
                t = dump2bg( raw_cover_cd, 
                             bg_fh_top_strand_cov, 
                             bg_fh_bot_strand_cov, 
                             b_ref, # current boundary ref
                             b_coord, # current boundary coord
                              )
                #self.stat_dict['n_of_nt_covered'] += n
                
                chunk_size = 0
                print('{0} nucleotides with coverage'.format(str(self.stat_dict['n_of_nt_covered'])))
                #TODO: fixme
                print('{0} sequenced bases'.format(str(self.stat_dict['n_of_nt_covered'])))
                continue
        
        #TODO: perform final dump
        b_ref = d['RNAME'] # current ref is the boundary ref
        b_coord = d['AEND'] # the final boundary is the end coord
        
        print('final vector dump') #debugging
        t = dump2bg( cleavage_cd,
                     bg_fh_top_strand_term, 
                     bg_fh_bot_strand_term, 
                     b_ref, # current boundary ref
                     b_coord, # current boundary coord
                     rate_mode = cleav_rate_mode, 
                     rate_cutoff = cleav_rate_cutoff,
                     rate_denom_cd = raw_cover_cd,
                      )
        #self.stat_dict['n_of_nt_termini'] += n
        
        t = dump2bg( oneD_cd,
                     bg_fh_top_strand_1D, 
                     bg_fh_bot_strand_1D, 
                     b_ref, # current boundary ref
                     b_coord, # current boundary coord
                     rate_mode = oneD_rate_mode, 
                     rate_cutoff = oneD_rate_cutoff,
                     rate_denom_cd = raw_cover_cd,
                      )
        #self.stat_dict['n_of_oneD_operations'] += n
        
        # The raw coverage ome_dict must be dumped last because it 
        # serves as the denominator for the rate calculations for 
        # cleavage and oneD
        t = dump2bg( raw_cover_cd,
                     bg_fh_top_strand_cov, 
                     bg_fh_bot_strand_cov, 
                     b_ref, # current boundary ref
                     b_coord, # current boundary coord
                      )
        #self.stat_dict['n_of_nt_covered'] += n
        
        chunk_size = 0
        print('{0} nucleotides with coverage'.format(str(self.stat_dict['n_of_nt_covered'])))
        #TODO: fixme
        print('{0} sequenced bases'.format(str(self.stat_dict['n_of_nt_covered'])))
        
        return

